module("L_HoneywellLib_emea", package.seeall)

--------------------------------------------------------------------------------
--
-- Honeywell Total Connect Comfort backend for EMEA V1 API. This support is taken
-- from work in the Evohome Web Service.
--
-- Creator:  Mikee

-- Version 1.0: Dec 27, 2015 Mikee
--   Initial non working cut. Will need an account to further the development.
--
-- Version 1.1: Feb 2, 2016 Mikee
--   Fix heat status

local main = { }

-- Upcalls
local debug = nil
local trace = nil
local log = nil
local err = nil
local log_http = nil
local updateCookies = nil
local createCookie = nil

local json = require("L_HoneywellTCC_json")
local ssl = require "ssl" -- rigpapa 2020-02-14
local https = require("ssl.https")
local K = require("L_HoneywellTCC_const")

--------------------------------------------------------------------------------
-- Constants

-- various constants
local TCC_SID = "urn:joeyd-com:serviceId:HoneywellTCC1"
local PROP_REAUTH_INTERVAL = "ReAuthInterval"

-- Honeywell Web portal constants
--local K_HOST = "rs.alarmnet.com"
--local K_AUTH_URL = "https://" .. K_HOST .. ":443/TotalConnectComfort/Auth/OAuth/Token"
--local K_EMEA_URL = "https://" .. K_HOST .. ":443/TotalConnectComfort/WebAPI/emea/api/v1"
local K_HOST = "tccna.honeywell.com"
local K_AUTH_URL = "https://" .. K_HOST .. "/Auth/OAuth/Token"
local K_EMEA_URL = "https://" .. K_HOST .. "/WebAPI/emea/api/v1"


local K_USER_AGENT = "Apache-HttpClient/UNAVAILABLE (java 1.4)"
local K_AUTHORIZATION = "Basic YjAxM2FhMjYtOTcyNC00ZGJkLTg4OTctMDQ4YjlhYWRhMjQ5OnRlc3Q="
local K_SCOPE = "EMEA-V1-Basic EMEA-V1-Anonymous EMEA-V1-Get-Current-User-Account"
local K_APP_ID = "b013aa26-9724-4dbd-8897-048b9aada249"

local LUASEC_PROTO = (ssl._VERSION or "0.5"):match("^0.5") and "tlsv1" or "any" -- rigpapa 2020-02-14

--------------------------------------------------------------------------------
-- Globals
local Token = nil
local TccDevice = 0
local RetryLimit = 1
local Cookies = { }
local AccountInfo = { }
local InstallationInfo = { }
local Locations = { }

--------------------------------------------------------------------------------
-- Local functions

--++
-- Make an http request with retry
--
-- @param spec_options - table of additional or override options
-- @return res -
--         code - https result code
--         header - response headers
--
function https_request(spec_options)
   trace("https_request", { url = spec_options.url })

   local res, code, header
   local base_options = {
      ["headers"] = {
     ["Accept"] = "application/json, application/xml, text/json, text/x-json, text/javascript, text/xml",
     ["Accept-Encoding"] = "sdch",
     ["Accept-Language"] = "en-US",
     ["Connection"] = "keep-alive",
--
--   ["Cache-Control"] ='no-store no-cache',
     ["Cache-Control"] ='no-cache',
--
     ["Host"] = K_HOST,
      },
   }
   if Token ~= nil then
      base_options.headers.Authorization = "bearer " .. Token
-- what is this ??
      base_options.headers.applicationId = K_APP_ID
   end


   for i=1,RetryLimit do
      local options = { }
      options.headers = { }
      for name,value in pairs(base_options) do
     if name == "headers" then
        for hname, hvalue in pairs(value) do
           options.headers[hname] = hvalue
        end
     else
        options[name] = value
     end
      end
      for name,value in pairs(spec_options) do
     if name == "headers" then
        for hname, hvalue in pairs(value) do
           options.headers[hname] = hvalue
        end
     else
        options[name] = value
     end
      end

      local cookie = createCookie(Cookies)
      if cookie ~= "" then
     options.headers["Cookie"] = cookie
      end

      -- Fix for LuaSec 0.8 and up, rigpapa 2020-02-14
      options.protocol = LUASEC_PROTO

      log_http("https req", { options=options })
      res, code, header = https.request(options)
      log_http("https resp", { res=res, code=code, header=header })

      if code ~= 400 and code ~= 504 then
     break
      end
      trace("https_request retry", { res=res, code=code })
   end

   trace("https_request return", { res=res, code=code })
   return res, code, header
end

--++
-- Login to the Honeywell Total Connect Comfort website
--
-- @param user - username
-- @param pass - password
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         token - session token
-- ------------------------------------------------------------------------------
local function login(user, pass)
   local time = os.time();

   local data = {
--      ["Content-Type"] = 'application/x-www-form-urlencoded; charset=utf-8',
--      ["Host"] =  'rs.alarmnet.com/',
--      ["Cache-Control"] ='no-store no-cache',
--      ["Pragma"] =    'no-cache',
      ["grant_type"] =  'password',
      ["scope"] =   'EMEA-V1-Basic EMEA-V1-Anonymous EMEA-V1-Get-Current-User-Account',
      ["Username"] =    user,
      ["Password"] =    pass,
-- where is client_id ??
--      ["Connection"] =    'Keep-Alive'
   }

   local body = lua_to_url(data)
   debug("body = " .. body)

   local result = { }
   local res, code, header = https_request {
      ["method"] = "POST",
      ["url"] = K_AUTH_URL,
      ["headers"] = {
     ["Pragma"] = 'no-cache',
     ["Authorization"] = K_AUTHORIZATION,
     ["Content-Type"] = "application/x-www-form-urlencoded; charset=utf-8",
     ["content-length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   last_login_attempt = os.time();

   local status, reason
   if (code ~= 200) then
      reason = "Login failed, code = " .. code
      if code == 401 then
     status = K.STATUS_AUTH_FAIL
      else
     status = K.STATUS_FAIL
      end
   else
      res = json.decode(table.concat(result))
      log("access res", { res=res })
      Token = res.access_token
      local refreshTime = res.expires_in

      status = K.STATUS_OK
      reason = "Success"

      log("login res", { token=Token })

      log("Schedule login refresh in " .. tostring(refreshTime - 30) .. " secs")
      luup.variable_set(TCC_SID, PROP_REAUTH_INTERVAL, refreshTime - 30, TccDevice)
   end

   return status, reason, Token
end

--++
-- Get the list of devices per location
--
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         locations - table of locations (See GetLocations for description)
--
local function getLocations()
   trace("getLocations")

   local devices = { }
   local status = K.STATUS_FAIL
   local reason = "Unknown failure"

   local result = { }
   local res, code, header = https_request {
      ["method"] = "GET",
      ["url"] = K_EMEA_URL .. "/userAccount",
      ["sink"] = ltn12.sink.table(result),
   }

   if code ~= 200 then
      if code == 401 then
     status = K.STATUS_AUTH_FAIL
     reason = "Authorization failed"
      else
     status = K.STATUS_HTTP_FAIL
     reason = "Unexpected HTTP failure, code " .. code
      end
   else
      debug("result", { result=result })
      res = json.decode(table.concat(result))
      log("access res", { res=res })
      AccountInfo = res
      local userId = AccountInfo.userId

      local result = { }
      local res, code, header = https_request {
     ["method"] = "GET",
     ["url"] = K_EMEA_URL .. "/location/installationInfo?userId=" .. userId .. "&includeTemperatureControlSystems=True",
     ["sink"] = ltn12.sink.table(result),
      }

      if code ~= 200 then
     if code == 401 then
        status = K.STATUS_AUTH_FAIL
        reason = "Authorization failed"
     else
        status = K.STATUS_HTTP_FAIL
        reason = "Unexpected HTTP failure, code " .. code
     end
      else
     debug("result", { result=result })
     res = json.decode(table.concat(result))
     log("access res", { res=res })
     InstallationInfo = res

     for _,loc in ipairs(InstallationInfo) do
        local result = { }
        local res, code, header = https_request {
           ["method"] = "GET",
           ["url"] = K_EMEA_URL .. "/location/" .. loc.locationInfo.locationId .. "/status?includeTemperatureControlSystems=True",
           ["sink"] = ltn12.sink.table(result),
        }

        if code ~= 200 then
           if code == 401 then
          status = K.STATUS_AUTH_FAIL
          reason = "Authorization failed"
           else
          status = K.STATUS_HTTP_FAIL
          reason = "Unexpected HTTP failure, code " .. code
           end
           break
        else
           status = K.STATUS_OK
           reason = "Success"

           debug("result", { result=result })
           res = json.decode(table.concat(result))
           log("access res", { res=res })

           local tbl = { }
           for _,gw in ipairs(res.gateways) do
          log("gateway", { gw = gw })
          for _,tcs in ipairs(gw.temperatureControlSystems) do
             log("system", { tcs = tcs })
             for _,zone in ipairs(tcs.zones) do
            log("zone", { zone = zone })
            tbl[loc.locationInfo.locationId .. ":" .. zone.zoneId] = { Name = zone.name }
             end
          end
           end
           devices[loc.locationInfo.locationId] = { Name = loc.locationInfo.name, Devices = tbl }
        end
     end
      end
   end

   trace("getLocations return", { status=status, reason=reason })
   return status, reason, devices
end

--++
-- Convert the response for systems and zones to TCC status
--
local function convertStatus(zone, system)
   local res = { }
   res.Success = true
   res.Load = 0

   -- Current temp and units
   res.DisplayUnits = "C"
   res.DisplayTemp = zone.temperatureStatus.temperature

   -- Heat status
   if zone.heatSetpointStatus ~= nil then
      res.HeatSetpoint = zone.heatSetpointStatus.targetTemperature
      if zone.heatSetpointStatus.setpointMode == "PermanentOverride" then
     res.StatusHeat = 2
      elseif zone.heatSetpointStatus.setpointMode == "TemporaryOverride" then
     res.StatusHeat = 1
      else
     res.StatusHeat = 0
      end
   else
      res.HeatSetpoint = 0
      res.StatusHeat = 0
   end
   res.SchedHeatSetpoint = 0
   res.HeatNextPeriod = 0

   -- Cool status
   if zone.coolSetpointStatus ~= nil then
      res.CoolSetpoint = zone.coolSetpointStatus.targetTemperature
      if zone.coolSetpointStatus.setpointMode == "PermanentOverride" then
     res.StatusCool = 2
      elseif zone.coolSetpointStatus.setpointMode == "TemporaryOverride" then
     res.StatusCool = 1
      else
     res.StatusCool = 0
      end
   else
      res.CoolSetpoint = 0
      res.StatusCool = 0
   end
   res.SchedCoolSetpoint = 0
   res.CoolNextPeriod = 0

   -- System status
   -- Normal/Auto = 0, HeatingOff = 1, Economy = 2, Away = 3, DayOff = 4, Custom = 6
   if system.systemModeStatus.mode == "HeatingOff" then
      res.SystemSwitchPosition = K.SYSTEM_SWITCH_OFF_ID;
   else
      res.SystemSwitchPosition = K.SYSTEM_SWITCH_HEAT_ID;
   end
   return res
end

--++
-- Get the status of a thermostat
--
-- @param device_id = TCC device id
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         stat - table of thermostat data (see GetStatus for description)
--
local function getStatus(device)
   trace("getStatus", { device=device })

   local locationId, deviceId = string.match(device, "([^:]-):(.*)")
   if locationId == nil or deviceId == nil then
      return K.STATUS_FAIL, "Invalid deviceId", nil
   end

   local status = K.STATUS_FAIL
   local reason = "Unknown failure"
   local data = nil

   local result = { }
   local res, code, header = https_request {
      ["method"] = "GET",
      ["url"] = K_EMEA_URL .. "/location/" .. locationId .. "/status?includeTemperatureControlSystems=True",
      ["sink"] = ltn12.sink.table(result),
   }

   if code ~= 200 then
      if code == 401 then
     status = K.STATUS_AUTH_FAIL
     reason = "Authorization failed"
      else
     status = K.STATUS_HTTP_FAIL
     reason = "Unexpected HTTP failure, code " .. code
      end
   else
      status = K.STATUS_OK
      reason = "Success"

      debug("result", { result=result })
      res = json.decode(table.concat(result))
      log("access res", { res=res })

      local loc_info = { }
      for _,gw in ipairs(res.gateways) do
     local gw_info = { }
     log("gateway", { gw = gw })
     for _,tcs in ipairs(gw.temperatureControlSystems) do
        local tcs_info = { }
        log("system", { tcs = tcs })
        for _,zone in ipairs(tcs.zones) do
           tcs_info[zone.zoneId] = zone.name;
           log("zone", { zone = zone })
           if zone.zoneId == deviceId then
          data = convertStatus(zone, tcs)
          status = K.STATUS_OK
          reason = "Success"
          break
           end
        end
        gw_info[tcs.systemId] = tcs_info;
     end
     loc_info[gw.gatewayId] = gw_info;
      end
      Locations[res.locationId] = loc_info;
   end

   trace("getStatus return", { status=status, reason=reason })
   return status, reason, data
end

--++
-- Obtain schedule from the website for specified thermostat
--
-- @param device_id - TCC device id
--
local function get_schedule(device_id)
   return K.STATUS_FAIL, "Not implemented", nil
end

--++
-- Set thermostat values
--
-- @param device - TCC device ID
-- @param req - Requested state change (See SetStatus for description
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--
local function setStatus(device, req)
   trace("setStatus", { device=device, req=req })

   local locationId, deviceId = string.match(device, "([^:]-):(.*)")
   if locationId == nil or deviceId == nil then
      return K.STATUS_FAIL, "Invalid device", nil
   end

   if req.HeatSetpoint == nil or req.StatusHeat == nil then
      return K.STATUS_OK, "Nothing to do", nil
   end

   local status = K.STATUS_FAIL
   local reason = "Unknown failure"

   -- XXX convert HeatNextPeriod to time
   local body = {}
   body["HeatSetpointValue"] = req.HeatSetpoint
   if req.StatusHeat == K.STATUS_TYPE_TEMP_ID then
      -- Note these values are backwards from the other APIs, 1=perm, 2=temp
      body["SetpointMode"] = 2
      body["TimeUntil"] = os.date("!%Y-%m-%dT%H:%M:%SZ", os.time() + 4*60*60) -- hold for 4 hours
   elseif req.StatusHeat == K.STATUS_TYPE_PERM_ID then
      -- Note these values are backwards from the other APIs, 1=perm, 2=temp
      body["SetpointMode"] = 1
      body["TimeUntil"] = json.null
   end

   local body = json.encode(body)
   log("data", { body=body })

   local result = { }
   local res, code, header = https_request {
      ["method"] = "PUT",
      ["url"] = K_EMEA_URL .. "/temperatureZone/" .. deviceId .. "/heatSetpoint",
      ["headers"] = {
     ["Content-Type"] = "application/json",
     ["Content-Length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   debug("result", { result=result })
   res = json.decode(table.concat(result))
   log("access res", { res=res })

   if code == 200 or code == 201 then
      status = K.STATUS_OK
      reason = "Success"
   elseif code == 401 then
      status = K.STATUS_AUTH_FAIL
      reason = "Authorization failed"
   end

   trace("setStatus", { status=status, reason=reason })
   return status, reason
end

--++
-- Set system mode
--
-- @param device - TCC device ID
-- @param req - Requested state change (See SetStatus for description
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--
local function setMode(device, req)
   trace("setStatus", { device=device, req=req })

   local locationId, deviceId = string.match(device, "([^:]-):(.*)")
   if locationId == nil or deviceId == nil then
      return K.STATUS_FAIL, "Invalid device", nil
   end

   if req.SystemSwitch == nil then
      return K.STATUS_OK, "Nothing to do", nil
   end

   -- Find the device and its parent control system
   local tcsId = nil
   for loc_id, loc_info in pairs(Locations) do
      log("loc_id", { loc_id=loc_id, locationId=locationId })
      if loc_id == locationId then
     for gw_id,gw_info in pairs(loc_info) do
        log("gw_id", { gw_id=gw_id })
        for tcs_id, tcs_info in pairs(gw_info) do
           log("tcs_id", { tcs_id=tcs_id })
           for zone_id, zone_name in pairs(tcs_info) do
          log("zone_id", { zone_id=zone_id, deviceId=deviceId })
          if zone_id == deviceId then
             tcsId = tcs_id
             break;
          end
           end
        end
     end
      end
   end

   if tcsId == nil then
      log("locations", { locs=Locations });
      return K.STATUS_FAIL, "Unknown device", nil
   end

   local status = K.STATUS_FAIL
   local reason = "Unknown failure"

   local args = {}
   args["TimeUntil"] = json.null
   args["Permanent"] = true
--   args["TimeUntil"] = os.date("!%Y-%m-%dT%H:%M:%SZ", os.time() + 4*60*60) -- hold for 4 hours
--   args["Permanent"] = false
   if req.SystemSwitch == K.SYSTEM_SWITCH_HEAT_ID then
      args["SystemMode"] = 0 -- Normal
   elseif req.SystemSwitch == K.SYSTEM_SWITCH_OFF_ID then
      args["SystemMode"] = 1 -- Off
   else
      args = "None"
   end

   local body = json.encode(args)
   log("data", { body=body })

   local result = { }
   local res, code, header = https_request {
      ["method"] = "PUT",
      ["url"] = K_EMEA_URL .. "/temperatureControlSystem/" .. tcsId .. "/mode",
      ["headers"] = {
     ["Content-Type"] = "application/json",
     ["Content-Length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   debug("result", { result=result })
   res = json.decode(table.concat(result))
   log("access res", { res=res })

   if code == 200 or code == 201 then
      status = K.STATUS_OK
      reason = "Success"
   elseif code == 401 then
      status = K.STATUS_AUTH_FAIL
      reason = "Authorization failed"
   end

   trace("setStatus", { status=status, reason=reason })
   return status, reason
end

--------------------------------------------------------------------------------
-- APIs

--++
-- Perform initialization
--
function Init(devId, imports)
   TccDevice = devId
   main = imports

   debug = main.log_http
   trace = main.trace
   log = main.log
   log_http = main.log_http
   err = main.err

   lua_to_url = main.lua_to_url
   updateCookies = main.updateCookies
   createCookie = main.createCookie
end

--++
-- Login and return whatever token is reuired for subsequent calls. This token
-- will be passed on all library calls
--
-- @param user - string username
-- @param pass - string password
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         token - token to use on subsequent calls
function Login(user, pass)
   trace("EMEA.Login", { user=user, pass=pass })

   local status, reason, session = login(user, pass)

   trace("EMEA.Login return", { status=status, reason=reason, session=session })
   return status, reason, session
end

--++
-- Return a structure of location/device information
--
-- @param token - authentication token
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         devices {
--                  { locationID = {
--                    Name = "",  - string location name
--                    Devices = {
--                     { deviceId = { Name = name }},...
--                    }
--                   },...
--                  }
function GetDevices(token)
   trace("EMEA.GetDevices", { token=token })

   Token = token
   if Token == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, devices = getLocations()

   trace("EMEA.GetDevices return", { status=status, reason=reason, devices=devices })
   return status, reason, devices
end

--++
-- Return the status of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         status - see L_HoneywellTCC_const for Get Status record
--
function GetStatus(token, deviceId)
   trace("EMEA.GetStatus", { token=token, deviceId=deviceId })

   Token = token
   if Token == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, info = getStatus(deviceId)

   trace("EMEA.GetStatus return", { status=status, reason=reason, info=info })
   return status, reason, info
end

--++
-- Set the operating parameters of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @param req - see L_HoneywellTCC_const for Set Status record
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--
function SetStatus(token, deviceId, req)
   trace("EMEA.SetStatus", { token=token, deviceId=deviceId })

   Token = token
   if Token == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason = setStatus(deviceId, req)
   if status == K.STATUS_OK then
      status, reason = setMode(deviceId, req)
   end

   trace("EMEA.SetStatus return", { status=status, reason=reason })
   return status, reason
end
